@@ -1,9 +1,18 @@
-{-# LANGUAGE BangPatterns, CPP, FlexibleInstances, OverloadedStrings,
-             Rank2Types #-}
-#ifdef GENERICS
-{-# LANGUAGE DefaultSignatures, TypeOperators, KindSignatures, FlexibleContexts,
-             MultiParamTypeClasses, UndecidableInstances, ScopedTypeVariables #-}
-#endif
+{-# LANGUAGE
+    BangPatterns,
+    CPP,
+    DefaultSignatures,
+    FlexibleContexts,
+    FlexibleInstances,
+    KindSignatures,
+    MultiParamTypeClasses,
+    OverlappingInstances,
+    OverloadedStrings,
+    Rank2Types,
+    ScopedTypeVariables,
+    TypeOperators,
+    UndecidableInstances
+    #-}
 module Data.Csv.Conversion
     (
     -- * Type conversion
@@ -11,6 +20,7 @@ module Data.Csv.Conversion
     , FromRecord(..)
     , FromNamedRecord(..)
     , ToNamedRecord(..)
+    , DefaultOrdered(..)
     , FromField(..)
     , ToRecord(..)
     , ToField(..)
@@ -29,17 +39,20 @@ module Data.Csv.Conversion
     , (.=)
     , record
     , namedRecord
+    , header
     ) where
 
-import Control.Applicative (Alternative, (<$>), (<|>), empty)
+import Control.Applicative (Alternative, (<|>), empty)
 import Control.Monad (MonadPlus, mplus, mzero)
 import Data.Attoparsec.ByteString.Char8 (double)
 import qualified Data.Attoparsec.ByteString.Char8 as A8
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Char8 as B8
 import qualified Data.ByteString.Lazy as L
+import Data.Hashable (Hashable)
 import qualified Data.HashMap.Lazy as HM
 import Data.Int (Int8, Int16, Int32, Int64)
+import qualified Data.IntMap as IM
 import qualified Data.Map as M
 import qualified Data.Text as T
 import qualified Data.Text.Encoding as T
@@ -50,23 +63,19 @@ import qualified Data.Vector as V
 import qualified Data.Vector.Unboxed as U
 import Data.Word (Word8, Word16, Word32, Word64)
 import GHC.Float (double2Float)
+import GHC.Generics
 import Prelude hiding (lookup, takeWhile)
 
 import Data.Csv.Conversion.Internal
 import Data.Csv.Types
 
 #if !MIN_VERSION_base(4,8,0)
-import Control.Applicative (Applicative, (<*>), (<*), (*>), pure)
+import Control.Applicative (Applicative, (<$>), (<*>), (<*), (*>), pure)
 import Data.Monoid (Monoid, mappend, mempty)
 import Data.Traversable (traverse)
 import Data.Word (Word)
 #endif
 
-#ifdef GENERICS
-import GHC.Generics
-import qualified Data.IntMap as IM
-#endif
-
 ------------------------------------------------------------------------
 -- bytestring compatibility
 
@@ -113,10 +122,8 @@ fromStrict = L.fromChunks . (:[])
 class FromRecord a where
     parseRecord :: Record -> Parser a
 
-#ifdef GENERICS
     default parseRecord :: (Generic a, GFromRecord (Rep a)) => Record -> Parser a
     parseRecord r = to <$> gparseRecord r
-#endif
 
 -- | Haskell lacks a single-element tuple type, so if you CSV data
 -- with just one column you can use the 'Only' type to represent a
@@ -140,12 +147,11 @@ newtype Only a = Only {
 -- > John,56
 -- > Jane,55
 class ToRecord a where
+    -- | Convert a value to a record.
     toRecord :: a -> Record
 
-#ifdef GENERICS
     default toRecord :: (Generic a, GToRecord (Rep a) Field) => a -> Record
     toRecord = V.fromList . gtoRecord . from
-#endif
 
 instance FromField a => FromRecord (Only a) where
     parseRecord v
@@ -365,10 +371,8 @@ instance (ToField a, U.Unbox a) => ToRecord (U.Vector a) where
 class FromNamedRecord a where
     parseNamedRecord :: NamedRecord -> Parser a
 
-#ifdef GENERICS
     default parseNamedRecord :: (Generic a, GFromNamedRecord (Rep a)) => NamedRecord -> Parser a
     parseNamedRecord r = to <$> gparseNamedRecord r
-#endif
 
 -- | A type that can be converted to a single CSV record.
 --
@@ -380,26 +384,66 @@ class FromNamedRecord a where
 -- >     toNamedRecord (Person name age) = namedRecord [
 -- >         "name" .= name, "age" .= age]
 class ToNamedRecord a where
+    -- | Convert a value to a named record.
     toNamedRecord :: a -> NamedRecord
 
-#ifdef GENERICS
-    default toNamedRecord :: (Generic a, GToRecord (Rep a) (B.ByteString, B.ByteString)) => a -> NamedRecord
+    default toNamedRecord ::
+        (Generic a, GToRecord (Rep a) (B.ByteString, B.ByteString)) =>
+        a -> NamedRecord
     toNamedRecord = namedRecord . gtoRecord . from
-#endif
 
-instance FromField a => FromNamedRecord (M.Map B.ByteString a) where
+-- | A type that has a default field order when converted to CSV. This
+-- class lets you specify how to get the headers to use for a record
+-- type that's an instance of 'ToNamedRecord'.
+--
+-- To derive an instance, the type is required to only have one
+-- constructor and that constructor must have named fields (also known
+-- as selectors) for all fields.
+--
+-- Right: @data Foo = Foo { foo :: !Int }@
+--
+-- Wrong: @data Bar = Bar Int@
+--
+-- If you try to derive an instance using GHC generics and your type
+-- doesn't have named fields, you will get an error along the lines
+-- of:
+--
+-- > <interactive>:9:10:
+-- >     No instance for (DefaultOrdered (M1 S NoSelector (K1 R Char) ()))
+-- >       arising from a use of ‘Data.Csv.Conversion.$gdmheader’
+-- >     In the expression: Data.Csv.Conversion.$gdmheader
+-- >     In an equation for ‘header’:
+-- >         header = Data.Csv.Conversion.$gdmheader
+-- >     In the instance declaration for ‘DefaultOrdered Foo’
+--
+class DefaultOrdered a where
+    -- | The header order for this record. Should include the names
+    -- used in the 'NamedRecord' returned by 'toNamedRecord'. Pass
+    -- 'undefined' as the argument, together with a type annotation
+    -- e.g. @'headerOrder' ('undefined' :: MyRecord)@.
+    headerOrder :: a -> Header  -- TODO: Add Generic implementation
+
+    default headerOrder ::
+        (Generic a, GToNamedRecordHeader (Rep a)) =>
+        a -> Header
+    headerOrder = V.fromList. gtoNamedRecordHeader . from
+
+instance (FromField a, FromField b, Ord a) => FromNamedRecord (M.Map a b) where
     parseNamedRecord m = M.fromList <$>
-                         (traverse parseSnd $ HM.toList m)
-      where parseSnd (name, s) = (,) <$> pure name <*> parseField s
+                         (traverse parseBoth $ HM.toList m)
 
-instance ToField a => ToNamedRecord (M.Map B.ByteString a) where
-    toNamedRecord = HM.fromList . map (\ (k, v) -> (k, toField v)) . M.toList
+instance (ToField a, ToField b, Ord a) => ToNamedRecord (M.Map a b) where
+    toNamedRecord = HM.fromList . map (\ (k, v) -> (toField k, toField v)) . M.toList
 
-instance FromField a => FromNamedRecord (HM.HashMap B.ByteString a) where
-    parseNamedRecord m = traverse (\ s -> parseField s) m
+instance (Eq a, FromField a, FromField b, Hashable a) => FromNamedRecord (HM.HashMap a b) where
+    parseNamedRecord m = HM.fromList <$>
+                         (traverse parseBoth $ HM.toList m)
 
-instance ToField a => ToNamedRecord (HM.HashMap B.ByteString a) where
-    toNamedRecord = HM.map toField
+instance (Eq a, ToField a, ToField b, Hashable a) => ToNamedRecord (HM.HashMap a b) where
+    toNamedRecord = HM.fromList . map (\ (k, v) -> (toField k, toField v)) . HM.toList
+
+parseBoth :: (FromField a, FromField b) => (Field, Field) -> Parser (a, b)
+parseBoth (k, v) = (,) <$> parseField k <*> parseField v
 
 ------------------------------------------------------------------------
 -- Individual field conversion
@@ -768,6 +812,10 @@ record = V.fromList
 namedRecord :: [(B.ByteString, B.ByteString)] -> NamedRecord
 namedRecord = HM.fromList
 
+-- | Construct a header from a list of 'B.ByteString's.
+header :: [B.ByteString] -> Header
+header = V.fromList
+
 ------------------------------------------------------------------------
 -- Parser for converting records to data types
 
@@ -849,8 +897,6 @@ runParser p = unParser p left right
 ------------------------------------------------------------------------
 -- Generics
 
-#ifdef GENERICS
-
 class GFromRecord f where
     gparseRecord :: Record -> Parser (f p)
 
@@ -938,4 +984,41 @@ instance ToField a => GToRecord (K1 i a) Field where
 instance (ToField a, Selector s) => GToRecord (M1 S s (K1 i a)) (B.ByteString, B.ByteString) where
     gtoRecord m@(M1 (K1 a)) = [T.encodeUtf8 (T.pack (selName m)) .= toField a]
 
-#endif
+-- We statically fail on sum types and product types without selectors
+-- (field names).
+
+class GToNamedRecordHeader a
+  where
+    gtoNamedRecordHeader :: a p -> [Name]
+
+instance GToNamedRecordHeader U1
+  where
+    gtoNamedRecordHeader _ = []
+
+instance (GToNamedRecordHeader a, GToNamedRecordHeader b) =>
+         GToNamedRecordHeader (a :*: b)
+  where
+    gtoNamedRecordHeader _ = gtoNamedRecordHeader (undefined :: a p) ++
+                             gtoNamedRecordHeader (undefined :: b p)
+
+instance GToNamedRecordHeader a => GToNamedRecordHeader (M1 D c a)
+  where
+    gtoNamedRecordHeader _ = gtoNamedRecordHeader (undefined :: a p)
+
+instance GToNamedRecordHeader a => GToNamedRecordHeader (M1 C c a)
+  where
+    gtoNamedRecordHeader _ = gtoNamedRecordHeader (undefined :: a p)
+
+-- | Instance to ensure that you cannot derive DefaultOrdered for
+-- constructors without selectors.
+instance DefaultOrdered (M1 S NoSelector a ()) => GToNamedRecordHeader (M1 S NoSelector a)
+  where
+    gtoNamedRecordHeader _ =
+        error "You cannot derive DefaultOrdered for constructors without selectors."
+
+instance Selector s => GToNamedRecordHeader (M1 S s a)
+  where
+    gtoNamedRecordHeader m
+        | null name = error "Cannot derive DefaultOrdered for constructors without selectors"
+        | otherwise = [B8.pack (selName m)]
+      where name = selName m
